import os
from tm351_nb_utils import exclude_hidden_items

import click
import os
import re
import nbformat
from urllib.parse import quote
from pathlib import Path


def replace_img_tags(markdown_text):
    def img_to_myst(match):
        attributes = match.groupdict()
        src = attributes.get("src", "")
        alt = attributes.get("alt", "")
        height = attributes.get("height", "")
        width = attributes.get("width", "")

        # URL encode the src path
        encoded_src = quote(src)

        myst = f"```{{image}} {encoded_src}\n"
        if alt:
            myst += f":alt: {alt}\n"
        if height:
            myst += f":height: {height}\n"
        if width:
            myst += f":width: {width}\n"
        myst += "```"
        return myst

    markdown_text_ = markdown_text
    # Regex to match img tags with attributes in any order
    img_pattern = r'<img\s+(?P<attrs>(?:src=["\'][^"\']+["\']|alt=["\'][^"\']*["\']|height=["\'][^"\']+["\']|width=["\'][^"\']+["\']|\s+)+)[^>]*>'

    # First, find all img tags
    img_tags = re.finditer(img_pattern, markdown_text)

    # Process each img tag
    for img_tag in img_tags:
        attrs = img_tag.group("attrs")

        # Extract individual attributes
        src_match = re.search(r'src=["\']([^"\']+)["\']', attrs)
        alt_match = re.search(r'alt=["\']([^"\']*)["\']', attrs)
        height_match = re.search(r'height=["\']([^"\']+)["\']', attrs)
        width_match = re.search(r'width=["\']([^"\']+)["\']', attrs)

        # Create a dictionary with the extracted attributes
        attributes = {
            "src": src_match.group(1) if src_match else "",
            "alt": alt_match.group(1) if alt_match else "",
            "height": height_match.group(1) if height_match else "",
            "width": width_match.group(1) if width_match else "",
        }

        # Replace the img tag with MyST syntax
        markdown_text = markdown_text.replace(
            img_tag.group(0),
            img_to_myst(type("Match", (), {"groupdict": lambda: attributes})),
        )

    return markdown_text, markdown_text_ == markdown_text


def replace_anchor_tags(markdown_text):
    def anchor_to_markdown(match):
        href = match.group("href")
        text = match.group("text")
        return f"[{text}]({href})"

    markdown_text_ = markdown_text
    # Regex to match anchor tags
    anchor_pattern = r'<a\s+href=["\'](?P<href>[^"\']+)["\'][^>]*>(?P<text>.*?)</a>'
    markdown_text = re.sub(anchor_pattern, anchor_to_markdown, markdown_text)
    return markdown_text, markdown_text_ == markdown_text


def replace_tags(markdown_text):
    # Replace img tags
    markdown_text, changed_img = replace_img_tags(markdown_text)

    # Replace anchor tags
    markdown_text, changed_anchor = replace_anchor_tags(markdown_text)

    return markdown_text, any([changed_img, changed_anchor])


def _process(p):
    """Replace HTML tags with MyST equivalent:

    - <img /> with {image} block;
    - <a> anchor with []() md link

    """

    if p.is_file() and p.suffix == ".ipynb":
        # Read notebook
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)

        changed = False

        # Process each cell
        for cell in nb.cells:
            if cell.cell_type == "markdown":
                cell.source, changed_ = replace_tags(cell.source)
            changed = changed_ or changed
        if changed:
            # Write the modified notebook
            with p.open("w") as f:
                nbformat.write(nb, f)
            print(f"Updated {p}")

    else:
        pass
        #print(f"{p} is not a Jupyter notebook file.")


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
def html_tag_replacer(paths, recursive):
    """Replace HTML image and anchor tags with MyST equivalent."""

    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p)
        else:
            for p in nb_dir.iterdir():
                _process(p)



# # Check cell run status issues
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os
import re
import nbformat
from pathlib import Path


def _process(p, check_run, check_not_run):
    """Process cell run status indications.

        - code cell not run (no cell run index)
        - cells run out of order (based on cell run index)
        - cell run index != code cell number in notebook

    """

    # Via chatgpt
    def _has_non_comment_text(s):
        # Regular expression to remove single-line comments
        s_no_single_line_comments = re.sub(r"#.*", "", s)

        # Regular expression to remove multi-line comments (triple single quotes or double quotes)
        s_no_comments = re.sub(
            r'""".*?"""|\'\'\'.*?\'\'\'', "", s_no_single_line_comments, flags=re.DOTALL
        )

        # Check if there is any text left that is not just whitespace
        return bool(re.search(r"\S", s_no_comments)), s_no_comments.strip()

    def _get_reports(p, nb, check_run, check_not_run):
        reports = []
        _i = 0  # Code cell count
        _run_ix_list = []
        for i, cell in enumerate(nb["cells"]):
            if cell["cell_type"] == "code":
                _i = _i + 1
            else:
                continue
            _x = -999
            if "execution_count" in cell:
                _x = cell["execution_count"]
                _run_ix_list.append((_i, _x))
            if check_run:
                if _x is None:
                    _report = f"Code cell {_i}: not executed"
                    reports.append(_report)
                elif _x != _i:
                    _report = f"Code cell {_i}: execute count mismatch {_x}!={_i}"
                    reports.append(_report)
            if check_not_run:
                if _x is not None:
                    _report = f"Code cell {_i} has been executed: {_i}/{_x}"
                    reports.append(_report)
                if "outputs" in cell and cell["outputs"]:
                    _report = f"Code cell {_i} has been executed — cell outputs on {_i}/{_x}"
                    reports.append(_report)
         
            source = cell.get("source", "")
            # Cope with v4 and v5 notebooks
            if isinstance(source, list):
                source = "".join(source).strip()
            if not source:
                _report = f"Code cell {_i}({i})/{_x} is empty"
                reports.append(_report)
            else:
                actual_code = _has_non_comment_text(source)
                if not actual_code[0]:
                    _report = f"Code cell {_i}({i})/{_x} only contains comments"
                    reports.append(_report)
            if "outputs" in cell:
                for output in cell["outputs"]:
                    if "name" in output and output["name"] == "stderr":
                        msg = output["text"].split("\n")[0]
                        _report = f"Warning in cell {_i}/{_x}: {msg}"
                        reports.append(_report)
                    elif "output_type" in output and output["output_type"] == "error":
                        msg = output["ename"] if "ename" in output else "???"
                        _report = f"Warning in cell {_i}({i})/{_x}: {msg}"
                        reports.append(_report)

        # Get actually executed cells
        executed_cells = [c for c in _run_ix_list if c[1] is not None]
        # Check ascending order
        out_of_order_pairs = []
        for i in range(len(executed_cells) - 1):
            if executed_cells[i][1] >= executed_cells[i + 1][1]:
                out_of_order_pairs.append((executed_cells[i], executed_cells[i + 1]))

        if out_of_order_pairs:
            for oop in out_of_order_pairs:
                reports.append(f"Out of order cells: code cell {oop[1]} follows {oop[0]} ")
        return reports

    fixes = []

    if p.is_file() and p.suffix == ".ipynb":
        # Read notebook
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)
            _fixes = _get_reports(p, nb, check_run, check_not_run)
            #if not _fixes:
            #    _fixes = ["No cell run status issues"]
            if _fixes:
                fixes.append(f"\nCell run status checks for {p}:\n\t"+"\n\t".join(_fixes)+"\n")

        if fixes:
            print("\n".join(fixes)) 


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
@click.option("--check-run", "-r", is_flag=True, help="Check if cells run.")
@click.option("--check-not-run", "-N", is_flag=True, help="Check if cells are not run.")
def cell_run_status_checker(paths, recursive, check_run, check_not_run):
    """Check cell run status indications."""

    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir, check_run, check_not_run)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p, check_run, check_not_run)
        else:
            for p in nb_dir.iterdir():
                _process(p, check_run, check_not_run)



import os
import jupytext
from pathlib import Path
import io

import codecs


def _dir_file_handler(path, _f, filetype='.md'):
    """Get the filename for a single file on a specified path."""
    f = os.path.join(path, _f)
    if f.endswith(filetype):
        return f
    return None

# TO DO - at the moment assumes markdown
# Need to accommodate notebook
# One approach might be to:
# - accept differnt text inputs (various markdown flavours, ipynb etc)
def detect_nb_doctype(path):
    """Dectect what sort of document we are dealing with."""
    # This is inefficient because we end up opening the file twice?
    with io.open(fp, encoding="utf-8") as f:
        txt = f.read()
        typ = jupytext.formats.divine_format(txt)
    return typ
# - cast to notebook with Jupytext
# - process notebook
# - rewrite output with same flavour as input
def _dir_handler(path):
    """Handle all the notebooks in a specific directory."""
    filelist=[]
    for _f in os.listdir(path):
        fn = _dir_file_handler(path, _f)
        if fn:
            filelist.append(fn)
    return filelist

def _path_full_processor(path):
    """Process file path."""
    for fn in _dir_handler(path):
        # Read file
        with open(fn, 'r') as f:
            # Process File
            txt = f.read()
            txt = process_headers(txt)
            txt = process_typos(txt, word_replacement, case_replacement)

        with open(fn, 'w') as f:
            # Write File
            f.write(txt)


def detect_encoding(file_path):
    encodings = ["utf-8", "ascii", "latin-1", "utf-16", "utf-32"]
    for encoding in encodings:
        print(f"trying to open file as {encoding}")
        try:
            with codecs.open(file_path, "r", encoding=encoding) as file:
                file.read()
            print(f"okay... opened with {encoding}")
            return encoding
        except UnicodeDecodeError:
            print(f"hmm... {encoding} gave UnicodeDecodeError")
            continue
    return None


def defensive_open(p, mode="r"):
    """Open a file with detected encoding or default to 'utf-8'."""
    detected_encoding = detect_encoding(p)
    encoding = "utf-8" if not detected_encoding else detected_encoding
    return p.open(mode, encoding=encoding)



# # Simple Routine to remove blank cells,
# either throughout or just at end of notebook
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path

def _process(p, all, cell_typs, stripwhitespace):
    """Clean empty cells."""

    # Function to check if a cell is empty and of an approved type
    def is_empty_approved(cell):
        source = cell.get("source", "")
        # Cope with v4 and v5 notebooks
        if isinstance(source, list):
            source = "".join(source)

        if not stripwhitespace:
            return cell["cell_type"] in cell_typs and not source
        return cell["cell_type"] in cell_typs and not source.strip()

    if p.is_file() and p.suffix == ".ipynb":
        # Read notebook
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)
            cell_count = len(nb.cells)
            if all:
                nb.cells = [cell for cell in nb.cells if not is_empty_approved(cell)]
            else:
                # Strip empty cells from the start
                while nb.cells and is_empty_approved(nb.cells[0]):
                    nb.cells.pop(0)

                # Strip empty cells from the end
                while nb.cells and is_empty_approved(nb.cells[-1]):
                    nb.cells.pop()

        if cell_count != len(nb.cells):
            print(f"Updating {p}")
            # By default, we do not convert notebook version
            nbformat.write(nb, p.open("w"), nbformat.NO_CONVERT)


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
@click.option(
    "--all/--no-all",
    default=False,
    help="Remove all empty cells, not just empty initial and final cells.",
)
@click.option(
    "--blitz/--no-blitz",
    default=True,
    help="Review all cell types (false if any explicit cell types are set)",
)
@click.option(
    "--code/--no-code",
    default=False,
    help="Check code cell.",
)
@click.option(
    "--md/--no-md",
    default=False,
    help="Check markdown cell.",
)
@click.option(
    "--raw/--no-raw",
    default=False,
    help="Check raw cell.",
)
@click.option(
    "--stripwhitespace/--no-stripwhitespace",
    default=True,
    help="Strip whitespace.",
)
def empty_cell_cleaner(paths, recursive, all, blitz, code, md, raw, stripwhitespace):
    """Clean empty cells."""
    cell_typs = set()
    if blitz:
        cell_typs = {"code", "markdown", "raw"}
    if code or md or raw:
        cell_typs = set()
    if code:
        cell_typs.add("code")  
    if md:
        cell_typs.add("markdown")  
    if raw:
        cell_typs.add("raw")  
    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir, all, cell_typs, stripwhitespace)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p, all, cell_typs, stripwhitespace)
        else:
            for p in nb_dir.iterdir():
                _process(p, all, cell_typs, stripwhitespace)



# # Simple Routine to clear a ky from metadata
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path

def _process(p, keys):
    """Handle metadata key cleaner."""

    if p.is_file() and p.suffix == '.ipynb':
        updated = False

        # Read notebook
        with p.open('r') as f:
            #print(f"Trying {p} with {keys}")

            nb = nbformat.read(f, nbformat.NO_CONVERT)
            for _, cell in enumerate(nb['cells']):
                for key in keys:
                    if key in cell["metadata"]:
                        cell["metadata"].pop(key, None)
                        updated = True

        if updated:
            print(f"Updating {p}")
            nbformat.write(nb, p.open('w'), nbformat.NO_CONVERT)


@click.command()
@click.argument("path", type=click.Path(resolve_path=False))
@click.argument("keys", type=str, nargs=-1)
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
def cell_metadata_key_cleaner(path, keys, recursive):
    """Clean one or more metadata items by key from cell ."""
    keys = list(keys)
    # Parse notebooks
    nb_dir = Path(path)
    if nb_dir.is_file():
        _process(nb_dir, keys)
    elif recursive:
        exclude = set([])
        for dirname, subdirs, files in os.walk(path, topdown=True):
            subdirs[:] = [d for d in subdirs if d not in exclude]
            exclude_hidden_items(subdirs)
            for p in files:
                _process(Path(dirname) / p, keys)
    else:
        for p in nb_dir.iterdir():
            _process(p, keys)



import click
from pathlib import Path
import nbformat

# Update tags for empinken
# Need to update to use tags: style-activity, style-solution, style-learner style-tutor
# Update from tags: style_activity, style_solution, style_student style_commentate
# Update from metadata: 'commentate', 'activity', 'student', 'solution' true/false
@click.command()
@click.argument('path', type=click.Path(resolve_path=False))
def upgrade_empinken_tags(path='.'):
    """Upgrade tags."""
    def _upgrade_file(p):
        if p.is_file() and p.suffix == '.ipynb':
            print(f"Checking {p}")
            # Read notebook
            with p.open('r') as f:

                # parse notebook
                # nb = nbformat.read(f, as_version=nbformat.NO_CONVERT)
                # nb = nbformat.convert(nb, version)
                # opinionated
                nb = nbformat.read(f, as_version=version)
                # for each cell
                for i, _ in enumerate(nb['cells']):
                    metadata = nb['cells'][i]["metadata"]
                    if 'tags' not in metadata:
                        metadata['tags'] = []
                    tags = metadata['tags']
                    for typ in typs:
                        # Upgrade metadata . tutor|activity|learner|solution
                        if typ in metadata:
                            metadata.pop(typ, None)
                            tags.append(f'style-{new_typs[typ]}')

                        # Upgrade original tags format: style_activity|style_solution|style_student|style_commentate
                        if f'style_{typ}' in metadata.tags:
                            tags.remove(f'style_{typ}')
                            tags.append(f'style-{new_typs[typ]}')

                    # Dedupe
                    metadata['tags'] = list(set(tags))
                    if not metadata['tags']:
                        metadata.pop('tags', None)

                # Validate - exception if we fail
                nbformat.validate(nb)

                # Save notebook
                print(f"Updating {p}: {', '.join(tags)}")
                nbformat.write(nb, p.open('w'), version)

    # Opinionated on this
    version = 4 #nbformat.NO_CONVERT
    typs = ['commentate', 'activity', 'student', 'solution']
    new_typs = {"commentate":"tutor", "activity":"activity",
    "student":"learner", "solution":"solution"}

    # Iterate path
    nb_path = Path(path)
    if nb_path.is_file():
        _upgrade_file(nb_path)
    else:
        for p in nb_path.rglob("*"):  # nb_dir.iterdir():
            _upgrade_file(p)



import click
import nbformat
from pathlib import Path

@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
@click.option(
    "--hidden/--no-hidden", default=False, help="Include hidden files and directories."
)
def fix_collapse_headings_metadata(paths, recursive, hidden):
    """Fix collapsible headings metadata."""

    def _process(p):
        """Fix metadata in notebook cells."""

        if p.is_file() and p.suffix == ".ipynb":
            updated = False

            # Read notebook
            with p.open("r") as f:
                # print(f"Trying {p}")
                # heading_collapsed: true -> jp-MarkdownHeadingCollapsed: true
                nb = nbformat.read(f, nbformat.NO_CONVERT)
                for _, cell in enumerate(nb["cells"]):
                    if "heading_collapsed" in cell["metadata"]:
                        cell["metadata"]["jp-MarkdownHeadingCollapsed"] = cell[
                            "metadata"
                        ]["heading_collapsed"]
                        cell["metadata"].pop("heading_collapsed", None)
                        updated = True

            if updated:
                print(f"Fixing collapsible headings metadata in {p}")
                nbformat.write(nb, p.open("w"), nbformat.NO_CONVERT)

    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir)
        if recursive:
            for p in nb_dir.rglob("*"):
                if hidden:
                    _process(p)
                else:
                    if not any(part.startswith(".") for part in p.parts):
                        _process(p)

        else:
            for p in nb_dir.iterdir():
                _process(p)



# ou-tm351 - `nb_pub_utils`

# GOTCHA - Python on Mac logging in to Github: https://stackoverflow.com/a/42098127/454773

import click

import os
import shutil
import zipfile
import humanize
import datetime
import github
from tabulate import tabulate
from shlex import quote


import subprocess
def cli_command(cmd):
    try:
        out = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT).decode('utf-8')
    except subprocess.CalledProcessError as error:
        return (error.returncode, error.output.decode('utf-8'))
    
    if out!='': return (0, out)

def listify(item):
    ''' If presented with a string and a list is required, make a list... '''
    item = [] if item is None else item
    #We may be passed a tuple - in which case, listify...
    item = list(item) if isinstance(item,(list,tuple)) else [item]
    return item

def exclude_hidden_items(itemlist, exclude_hidden=True):
    ''' Exclude hidden items from ziplist '''
    if exclude_hidden:
        rmlist=[]
        for x in itemlist:
            if x.startswith('.'):
                rmlist.append(x)
        for x in rmlist:
            itemlist.remove(x)

def exclude_items(itemlist, excludes, exclude_hidden=True, ipynb_only=False):
    ''' Exclude items from ziplist '''

    for xd in set(itemlist).intersection(excludes):
        itemlist.remove(xd)

    if ipynb_only:
        for i in [_i for _i in itemlist if not _i.endswith("ipynb")]:
                itemlist.remove(i)
        
    if exclude_hidden: exclude_hidden_items(itemlist)


def notebookTest(path=None, filename=None, dir_excludes=None, file_excludes=None):
    ''' Run notebook tests over explicitly named files and directories.
    '''
    
    #Could probably define this recursively to handle mulitple paths/filenames...
    
    def pathmaker(path,filename):
        if not path or path in ['.']: return filename
        if not isinstance(path,list):
            return '/'.join([path,filename])
    
    sanitiser = """[regex1]
regex: <graphviz.files.Source at [^>]*>
replace: <graphviz.files.Source>

[regex2]
regex: CPU times: .*
replace: CPU times: CPUTIME

[regex3]
regex: Wall time: .*
replace: Wall time: WALLTIME

[regex4]
regex: .* per loop \(mean ± std. dev. of .* runs, .* loops each\)
replace: TIMEIT_REPORT
"""
    #tmp_fn = "_sanitise_cfg.cfg"
    #with open(tmp_fn, "w") as f:
    #    f.write(sanitiser)

    #cmd=f'py.test --nbval-sanitize-with {tmp_fn} '
    cmd=f'py.test '

    file_excludes = listify(file_excludes)

    for d in listify(dir_excludes):
        cmd = cmd + ' --ignore={} '.format(quote(d))
        print("*Not testing in directory: {}*".format(d))

    cmd = cmd+' --nbval '
    ## WARNING - TO DO - if we are running this from a notebook, also exclude path=='.'
    if path is None and filename is None:
        #Process current directory
        return cli_command(cmd)
    elif filename:
        #Process file(s) in directory
        if isinstance(filename, list):
            for _filename in filename:
                cmd = '{cmd} {filename}'.format(cmd=cmd, filename=pathmaker(path, quote(_filename)))
                resp=cli_command(cmd)
        else:
            cmd = '{cmd} {filename}'.format(cmd=cmd, filename=pathmaker(path, quote(filename)))
            resp=cli_command(cmd)
        return resp
    else:
        #Process files in path
        #If we pass a directory name in then the test will be run over all files in the directory
        #py.test accumulates the test responses
        resps = []
        for singlepath in listify(path):
            for dirname, subdirs, files in os.walk(singlepath):
                exclude_items(subdirs, dir_excludes)
                exclude_items(files, file_excludes, ipynb_only=True)
                print('Processing directory: {}'.format(dirname))
                with click.progressbar(files) as bar:
                    for filename in bar:
                        filepathname=os.path.join(dirname, filename)
                        cmd = '{cmd} {path}'.format(cmd=cmd, path=quote(filepathname))
                        resps.append( cli_command(cmd) )
        #for singlepath in listify(path):
        #    print("\nTesting in directory: {}".format(singlepath))
        #    if singlepath=='.':
        #        print('**DO NOT test in current directory from a notebook**')
        #    cmd = '{cmd} {path}'.format(cmd=cmd, path=quote(singlepath))
        #    resps.append( cli_command(cmd) )
        
        os.unlink(tmp_fn)
        return resps

def notebookProcessor(notebook, mode=None, outpath=None, outfile=None, inplace=True):
    ''' Clear notebook output cells.
    
        Process a single notebook, clearing cell outputs running cells until
        a warning, or running all cells despite warnings.
        
        Processed notebooks can be written to a specified directory or rendered inplace.
    '''
    
    if mode is None: return (-1, 'Mode not specified.')
    
    if outpath is not None and not os.path.exists(outpath):
        os.makedirs(outpath)
    
    if outfile is not None:
        outpath = '/'.join([outpath,outfile]) if outpath is not None else outfile
    
    cmd='jupyter nbconvert --to notebook'

    if mode in ['clearOutput', 'clearOutputTest' ]:
        cmd = '{cmd} --ClearOutputPreprocessor.enabled=True'.format(cmd=cmd)
    elif mode == 'run':
        cmd = '{cmd} --execute'.format(cmd=cmd)
    elif mode == 'runWithErrors':
        cmd = '{cmd} --ExecutePreprocessor.allow_errors=True --execute'.format(cmd=cmd)
    else: return (-1, 'Mode not specified correctly.')
    
    if outpath is None and inplace:
        cmd='{cmd} --inplace'.format(cmd=cmd)

    #Select file
    cmd='{cmd} {notebook}'.format(cmd=cmd,notebook=quote(notebook))
    
    #If output path not set, and --inplace is not set,
    #  nbformat will create a new file with same name ending: .nbformat.ipynb
    if outpath is not None:
        cmd ='{cmd} --output-dir {outpath}'.format(cmd=cmd, outpath=quote(outpath))
    #print(f"Trying command {cmd}")
    return cli_command(cmd)

def directoryProcessor(path,
                       mode=None, outpath=None, inplace=True,
                       include_hidden=False,
                       dir_excludes=None,
                       file_excludes=None, rmdir=False, currdir=False, subdirs=True,
                       reportlevel=1, logfile=None):
    ''' Process all the notebooks in one or more directories and
        (optionally) in associated subdirectories.
    
        Processed notebooks can be written to a specified directory or rendered inplace.
            
        Path hierarchies to notebooks in multiple directories or subdirectories are
        respected when writing to a specified output directory.
    '''
    
    def _process(outpath):
        ''' Process files associated with a particular directory '''
        processfiles=[f for f in files if f.endswith('.ipynb')]
        
        if subdirs:
            print(dirname)
            if outpath is not None:
                outpath='/'.join([outpath, dirname])
                if not os.path.exists(outpath):
                    os.makedirs(outpath)
        if not mode == 'tests':
            #print('About to process {}'.format(processfiles))
            with click.progressbar(processfiles) as bar:
                for filename in bar:
                    if not currdir and dirname=='.': continue
                    if reportlevel>1:
                        print("Processing >{}<".format('/'.join([dirname,filename])))
                    resp = notebookProcessor('/'.join([dirname,filename]), mode=mode, outpath=outpath, inplace=inplace )
                    if reportlevel>0 and resp and resp[0]!=0:
                        print("Error with {}".format('/'.join([dirname,filename])))
                    if logfile:
                        with open(logfile, "a") as out:
                            out.write(resp[1])

        #if mode in ['tests', 'clearOutputTest']:
        #    #Tests need to run in original dir in case of file dependencies
        #    testreport = notebookTest(path=dirname,dir_excludes=dir_excludes)
        #    print('tested:',dirname)
        #    print(testreport[1])
    
    #if mode == 'clearOutputTest':
    #    #If we are testing for warnings, need to test in original directory
    #    #  in case there are file dependencies
    #    outpath=None
    #    inplace=True
    
    if mode is None: return
    
    if isinstance(path, list):
        if rmdir:
            shutil.rmtree(outpath, ignore_errors=True)
            #Make sure we only delete the directory on the way in...
            rmdir=False
            
        for _path in path:
            #When provided with multiple directories, process each one separately
            #Note that subdirs for each directory can be handled automatically
            directoryProcessor(_path, mode, '/'.join([outpath, _path]), inplace,
                               include_hidden, dir_excludes, file_excludes,
                               rmdir, currdir, subdirs, reportlevel, logfile)
        return

    #TO DO - simplify this so we just pass one exclusion type then detect if file or dir?
    file_excludes = listify(file_excludes)
    dir_excludes = listify(dir_excludes)
    
    if outpath is not None and os.path.exists(outpath):
        if rmdir:
            print('\n***Deleting directory `{}` and all its contents....***\n\n'.format(outpath))
            shutil.rmtree(outpath, ignore_errors=True)
        else:
            print('\nOutput directory `{}` already exists. Remove it first by setting: rmdir=True\n'.format(outpath))
        
    #dir_excludes = [] if dir_excludes is None else dir_excludes 
    #file_excludes = [] if file_excludes is None else file_excludes
    if os.path.isfile(path):
        notebookProcessor(path, mode=mode, outpath=outpath, inplace=inplace )
    elif subdirs:
        for dirname, subdirs, files in os.walk(path):
            exclude_items(subdirs, dir_excludes, not include_hidden)
            exclude_items(files, file_excludes, not include_hidden)
            _process(outpath)
    # if passed a single file rather than directory path
    else:
        files=os.listdir(path)
        exclude_items(files, file_excludes, not include_hidden)
        dirname=path
        _process(outpath)

# Via Claude.ai
import tempfile
import warnings


def add_to_zipfile_with_warning(zf, filename, tmp_filename=None):
    try:
        # Temporarily set warnings to raise exceptions
        with warnings.catch_warnings():
            warnings.simplefilter("error", UserWarning)
            zf.write(filename, tmp_filename)
    except UserWarning as w:
        if "Duplicate name:" in str(w):
            print(f"Warning: Duplicate file in zip - {tmp_filename or filename}")
        else:
            # Re-raise other warnings
            raise


# via claude.ai
def process_notebook_with_temp_dir(filepathname, file_processor, zf):
    # Get the directory and filename of the notebook
    file_dir, filename = os.path.split(filepathname)

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Copy the original notebook to the temp directory
        temp_filepathname = os.path.join(temp_dir, filename)
        shutil.copy2(filepathname, temp_filepathname)

        if file_processor in ['clearOutput', 'runWithErrors'] and filename.endswith('.ipynb'):
            # Process the notebook in the temporary directory
            notebookProcessor(temp_filepathname, mode=file_processor, inplace=True)

        # Add the processed notebook to the zip file, preserving the original path
        # zf.write(temp_filepathname, filepathname)
        add_to_zipfile_with_warning(zf, temp_filepathname, filepathname)


def process_notebook_with_temp(filepathname, file_processor, zf):
    file_dir, filename = os.path.split(filepathname)

    # Create a temporary filename in the same directory
    with tempfile.NamedTemporaryFile(delete=False, dir=file_dir, suffix='.ipynb') as temp_file:
        temp_filepathname = temp_file.name
        try:
            # Copy the original notebook to the temporary filename
            shutil.copy2(filepathname, temp_filepathname)

            if file_processor in ['clearOutput', 'runWithErrors'] and filename.endswith('.ipynb'):
                # Process the notebook with the temporary filename
                notebookProcessor(temp_filepathname, mode=file_processor, inplace=True)

            # Add the processed notebook to the zip file, preserving the original path
            add_to_zipfile_with_warning(zf, temp_filepathname, filepathname)

        finally:
            # Ensure the temporary file is deleted after processing
            if os.path.exists(temp_filepathname):
                os.remove(temp_filepathname)

# Running zipper with a file_processor will change the cell state in current dir
# That is, notebooks are processed in place then zipped
# The notebooks as seen in the dir will reflect those in the zipfile
# We could modify this behaviour so it does not affect original notebooks?
def zipper(dirtozip, zipfilename,
           include_hidden=False,
           dir_excludes=None,
           file_excludes=None, 
           file_processor=None,
           reportlevel=1, rmdir=False, 
           zip_append=False):
    ''' Zip the contents of a directory and its subdirectories '''

    file_excludes = listify(file_excludes)
    dir_excludes = listify(dir_excludes)

    zip_permission = "a" if zip_append else "w"
    # Create a new/replacement zip file, rather than append if zipfile already exists
    zf = zipfile.ZipFile(zipfilename, zip_permission, compression=zipfile.ZIP_DEFLATED)

    # Don't zip files of same name as the zip file we are creating
    file_excludes.append(zipfilename)
    # if we have just a single file to zip and not a dir, zip that
    if os.path.isfile(dirtozip):
        if file_processor in [
            "clearOutput",
            "clearOutputTest",
            "runWithErrors",
        ] and dirtozip.endswith(".ipynb"):
            process_notebook_with_temp(dirtozip, file_processor, zf)
        # print("command is run, should be zipping now")
        else:
            zf.write(dirtozip)
    elif os.path.isdir(dirtozip):
        # https://stackoverflow.com/a/31779538/454773
        for dirname, subdirs, files in os.walk(dirtozip):
            exclude_items(subdirs, dir_excludes, not include_hidden)
            exclude_items(files, file_excludes, not include_hidden)
            print('Processing directory: {}'.format(dirname))
            zf.write(dirname)
            with click.progressbar(files) as bar:
                for filename in bar:
                    if reportlevel>1:print(filename)
                    filepathname=os.path.join(dirname, filename)
                    # There is no point using 'run': if there is an error, nbconvert will fail
                    if file_processor in ['clearOutput', 'runWithErrors'] and filename.endswith('.ipynb'):
                        # This introduces side effects - notebooks are processed in current path
                        process_notebook_with_temp(filepathname, file_processor, zf)
                        # notebookProcessor(filepathname, mode=file_processor, inplace=True)
                    else:
                        zf.write(filepathname)
    zf.close()

    # Is this too risky?!
    # if rmdir: shutil.rmtree(dirtozip, ignore_errors=True)
    return zipfilename

def insideZip(zfn, report=True):
    ''' Look inside a zip file.
        The report contains four columns: file_size, file compressed size, datetime and filename.
        Setting report=True returns a pretty printed report. '''
    if not os.path.isfile(zfn):
        print("\nHmm... {} doesn't seem to be a file?\n".format(zfn))
        return
    print('\nLooking inside zipfile: {}\n'.format(zfn))
    fz=zipfile.ZipFile(zfn)
    
    txt=[]
    for fn in fz.infolist():
        txt.append( [fn.file_size,
                     fn.compress_size,
                     datetime.datetime(*fn.date_time).isoformat(),
                     fn.filename] )
        print('{}, {}, {}, {}'.format(fn.file_size,
                     fn.compress_size,
                     datetime.datetime(*fn.date_time).isoformat(),
                     fn.filename))
    tabulate(txt, headers=['Full','Zip','Datetime','Path'],tablefmt="simple")
    return txt  


@click.command()
@click.option(
    "--file-processor", "-r", type=click.Choice(["clearOutput", "runWithErrors"])
)
@click.option("--include-hiddenfiles", "-H", is_flag=True, help="Include hidden files")
@click.option(
    "--exclude-dir",
    "-X",
    multiple=True,
    type=click.Path(resolve_path=False),
    help="Exclude specified directory",
)
@click.option(
    "--exclude-file",
    "-x",
    multiple=True,
    type=click.Path(resolve_path=False),
    help="Exclude specified file",
)
@click.option("--zip_append", "-a", is_flag=True, help="Add to existing zip file")
@click.option("--force", "-F", is_flag=True, default=False, help="Force the operation.")
@click.argument(
    "path", type=click.Path(resolve_path=False)
)  
# @click.argument('zipfile', type=click.File('wb'))
@click.argument("zipfile", type=click.Path())
def cli_zip(file_processor, include_hiddenfiles, exclude_dir, exclude_file, zip_append, force, path, zipfile):
    """Create a zip file from the contents of a specified directory.
    
    The zipper can optionally run a notebook processor on notebooks before zipping them to check that all cells are run or all cells are cleared.
    """
    print('You must be crazy using this...')

    if not zip_append and not force:
        print(f"\nThis would overwrite any previous {zipfile} file\n\tUse -F/--force to overwrite.\n\tUse -a, --zip_append to add to an existing zip file.")
        exit(-1)
    elif force and not zip_append:
        print(f"Deleting/completely overwriting {zipfile}")
    else:
        print(f"\nAppending zipped files to: {zipfile}\n")

    fn = zipper(path, zipfile,
                include_hidden=include_hiddenfiles,
                dir_excludes=exclude_dir,
                file_excludes=exclude_file, 
                file_processor=file_processor,
                zip_append=zip_append)

    print(f"\nZip file: {fn}\n")


# h/t Claude.ai
@click.command()
@click.argument("zipfile_path", type=click.Path(exists=True))
@click.argument("target_dir", type=click.Path(), default="zip_output")
def cli_unzip(zipfile_path, target_dir):
    """
    Unzip a file into a target directory.

    ZIPFILE_PATH: Path to the zip file to be extracted.
    TARGET_DIR: Directory to extract the contents to (default: zip_output).
    """
    try:
        # Ensure the target directory exists
        os.makedirs(target_dir, exist_ok=True)

        # Open and extract the zip file
        with zipfile.ZipFile(zipfile_path, "r") as zip_ref:
            zip_ref.extractall(target_dir)

        click.echo(f"Successfully extracted {zipfile_path} to {target_dir}")
    except zipfile.BadZipFile:
        click.echo(f"Error: {zipfile_path} is not a valid zip file", err=True)
    except PermissionError:
        click.echo(
            f"Error: Permission denied when trying to extract to {target_dir}", err=True
        )
    except Exception as e:
        click.echo(f"An unexpected error occurred: {str(e)}", err=True)


@click.command()
@click.option("--quiet", "-q", is_flag=True, help="Suppress the report.")
@click.option("--warnings", "-w", is_flag=True, help="Display warnings")
@click.argument("filename", type=click.Path(resolve_path=True), nargs=-1)
def cli_zipview(filename, warnings, quiet):
    """List the contents of one or more specified zipfiles.
    """
    zip_contents = []
    for f in listify(filename):
        zip_contents.append((f, insideZip(f)))

    if warnings and zip_contents:
        for (zn, item) in zip_contents:
            print(f"\n\n====== Zip file quality report: {zn} ======\n")
            for record in item:
                if record[1] > 1e6:
                    print(f"WARNING: \"{record[3]}\" looks quite large file ({humanize.naturalsize(record[0])} unzipped, {humanize.naturalsize(record[1])} compressed)")
                for _path in record[3].split('/'):
                    if len(_path) > 50:
                        print(f"ERROR: the filepath element \"{_path}\" in \"{record[3]}\" is too long (max. 50 chars)")
                    if _path.startswith("."):
                        print(f"WARNING: \"{record[3]}\" is a hidden file/directory (do you really need it in the zip file?)")
        print("\n===========================\n\n")

def _notebookTest(testitems,  outfile=None, dir_excludes=None, file_excludes=None):
    path=[]
    filename=[]
    
    for i in listify(testitems):
        if os.path.isdir(i):
            path.append(i)
        else:
            filename.append(i)
    resps = notebookTest(path=path, filename=filename, dir_excludes=dir_excludes, file_excludes=file_excludes)
    if isinstance(resps, tuple): resps = [resps]
    
    for resp in resps:
        if outfile:
            with open(outfile, "a") as out:
                out.write(resp[1])
            print('\nTest report written to {}'.format(outfile))
        else:
            print(resp[1])


@click.command()
@click.option('--exclude-dir','-X', multiple=True,type=click.Path(resolve_path=False), help='Do not recurse through specified directory when assembling tests.')
@click.option('--exclude-file','-x', multiple=True,type=click.Path(resolve_path=False), help='Exclude specified file')
@click.option('--outfile','-o', type=click.Path(resolve_path=False), help='Output report file. Leave this blank to display report on command line.')
@click.argument('testitems', type=click.Path(resolve_path=False),nargs=-1)
def cli_nbtest( exclude_dir, exclude_file, outfile, testitems):
    """Test specified notebooks and/or the notebooks in a specified directory or directories (`TESTITEMS`) using the `nbdime` plugin for `py.test`.
    
    Running `tm351nbtest` without any specified directory or file will assemble tests recursively from the current directory down."""
    testitems = testitems or '.'
    _notebookTest(testitems, outfile, exclude_dir, exclude_file)


@click.command()
@click.option('--file-processor','-r', type=click.Choice(['clearOutput', 'runWithErrors']), help='File processor actions that can be applied to notebooks using `nbconvert`')
@click.option('--outpath', '-O', type=click.Path(resolve_path=False), help='path to output directory')
@click.option('--inplace/--no-inplace',default=True, help='Run processors on notebooks inplace')
@click.option('--exclude-dir', '-X', multiple=True, type=click.Path(resolve_path=False), help='Exclude specified directory')
@click.option('--exclude-file','-x', multiple=True,type=click.Path(resolve_path=False), help='Exclude specified file')
@click.option('--include-hidden/--no-include-hidden',default=False, help='Include hidden files')
@click.option('--rmdir/--no-rmdir',default=False, help='Check the output directory is empty before we use it')
@click.option('--currdir/--no-currdir',default=False, help='Process files in current directory')
@click.option('--subdirs/--no-subdirs',default=True, help='Process files in subdirectories')
@click.option('--reportlevel', default=1, help='Reporting level')
@click.argument('path',type=click.Path(resolve_path=False))
def cli_nbrun(file_processor, outpath, inplace, exclude_dir, exclude_file, include_hidden, rmdir, currdir, subdirs, reportlevel, path):
    """Directory processor for notebooks - allows the user to run nbconvert operations on notebooks, such as running all cells or clearing all cells.
    
    To run tests, use: tm351nbtest
    
    To zip folders (with the option or running notebook processors on zipped files), use: tm351zip
    """
    directoryProcessor(path,
                       mode=file_processor, outpath=outpath, inplace=inplace,
                       include_hidden=include_hidden,
                       dir_excludes=exclude_dir,
                       file_excludes=exclude_file, rmdir=rmdir, currdir=currdir,
                       subdirs=subdirs,reportlevel=reportlevel)


from github import Github
import getpass

import base64
import logging
from github.GithubException import GithubException

def get_sha_for_tag(repository, tag):
    """
    Returns a commit PyGithub object for the specified repository and tag.
    """
    branches = repository.get_branches()
    matched_branches = [match for match in branches if match.name == tag]
    if matched_branches:
        return matched_branches[0].commit.sha

    tags = repository.get_tags()
    matched_tags = [match for match in tags if match.name == tag]
    if not matched_tags:
        raise ValueError('No Tag or Branch exists with that name')
    return matched_tags[0].commit.sha

def download_directory(repository, sha, server_path, outpath='gh_downloads', file_processor=None):
    """
    Download all contents at server_path with commit tag sha in
    the repository.
    """
    contents = repository.get_dir_contents(server_path, ref=sha)
    if not os.path.exists(outpath):
        os.makedirs(outpath)
            
    for content in contents:
        print("Downloading: %s" % content.path)
        if content.type == 'dir':
            download_directory(repository, sha, content.path, '/'.join([outpath,content.name]))
        else:
            try:
                path = content.path
                file_content = repository.get_contents(path, ref=sha)
                file_data = base64.b64decode(file_content.content)
                outpathfile='/'.join([outpath,content.name])
                file_out = open(outpathfile, "wb")
                file_out.write(file_data)
                file_out.close()
            except (IOError, github.GithubException) as exc:
                #If we fail over because of a large blog, use the data api for the download
                ret,error=exc.args
                if 'message' in error and error['message']=='Not Found':
                    print('Hmm... file not found? {}'.format(path))
                elif 'errors' in error and error['errors'][0]['code']=='too_large':
                    #print('...large file, trying blob download instead...')
                    file_content = repository.get_git_blob(content.sha)
                    file_data = base64.b64decode(file_content.content)
                    file_out = open('/'.join([outpath,content.name]), "wb")
                    file_out.write(file_data)
                    file_out.close()
                #logging.error('Error processing %s: %s', content.path, exc)
            #if content.name.endswith('.ipynb') and file_processor in ['clearOutput', 'clearOutputTest','runWithErrors' ]:
            #        notebookProcessor(outpathfile, file_processor)


def github_repo_branches(repository):
    return [br.name for br in repository.get_branches()]

def github_repo_topdirs(contents):
    return [i.name for i in contents if i.type=='dir']

DEFAULT_REPO='undercertainty/tm351'

@click.command()
@click.option('--github-user', '-u',  help="Your Github username.")
@click.option('--password', hide_input=True,
              confirmation_prompt=False)
@click.option('--repo','-r', prompt='Repository ({})'.format(DEFAULT_REPO),
              help='Repository name')
@click.option('--branch','-b',help='Branch or tag to download')
@click.option('--directory', help='Directory to download (or: all)')
@click.option('--savedir',type=click.Path(resolve_path=False),
              help='Directory to download repo / repo dir into; default is dir name')
@click.option('--file-processor', type=click.Choice(['clearOutput', 'runWithErrors']), help='Optionally specify a file processor to be run against downloaded notebooks.')
@click.option('--zip/--no-zip', default=False, help='Optionally create a zip file of the downloaded repository/directory with the same name as the repository/directory.')
@click.option('--auth/--no-auth', default=True, help="By default, run with auth (prompt for credentials)")
@click.option('--with-tests','-t',is_flag=True, help="Run tests on notebooks after download")
@click.option('--logfile',type=click.Path(resolve_path=False), help='Path to logfile')
def cli_gitrepos(github_user, password, repo, branch, directory, savedir, file_processor, zip, auth, with_tests, logfile):
    """Download files from a specified branch in a particular git repository.
    
    The download can also be limited to just the contents of a specified directory.
    
    Don't worry that there look to be a lot of arguments - you will be prompted for them if you just run: tm351gitrepos
    """
    
    if auth or github_user:
        if not github_user: github_user = click.prompt('\nGithub username')
        if not password: password = click.prompt('\nGithub password', hide_input=True)
        github = Github(github_user, password)
        #Show we're keeping no password...
        password = None
        auth = True
    else: github = Github()


    if auth:
        user = github.get_user()
        #organisations = github.get_user().get_orgs()
        print('Logging into git as {} ({})'.format(github_user, user.name))
    
    repo = repo or DEFAULT_REPO
    repository = github.get_repo(repo)

    if not branch:
        print('\nBranches available:\n\t{}'.format('\n\t'.join(github_repo_branches(repository)) ))
        branch = click.prompt('\nWhich branch? (master)')

    branch_or_tag_to_download = branch or 'master'
    sha = get_sha_for_tag(repository, branch_or_tag_to_download)
    
    another = ''
    while another!='-':
        if not directory:
            if branch!='master':
                contents = repository.get_dir_contents('.', ref=sha)
            else:
                contents = repository.get_dir_contents('.')
            print('\nYou can download all directories from this repo (all) or select one:\n\t{}'.format('\n\t'.join(github_repo_topdirs(contents))))
            directory = click.prompt('Which directory? (all)')

        directory_to_download = '.' if (not directory or directory=='all') else directory
        outpath = savedir or directory_to_download
        if outpath == '.' and savedir !='.': outpath=repo.replace('/','_')+'_files'
        
        msg='\nOkay... downloading {}/{}'.format(repo,directory_to_download ) 
        if file_processor is not None:
            msg = msg + ' using notebook processor: {}'.format(file_processor)
        else: msg = msg + ' with no notebook processing'
        print(msg)
        download_directory(repository, sha, directory_to_download, outpath,file_processor )

        if file_processor in ['clearOutput', 'clearOutputTest','runWithErrors' ]:
            click.echo('\nRunning notebook processor: {}'.format(file_processor))
            directoryProcessor(outpath, mode=file_processor, subdirs=True,
                                reportlevel=1, logfile=logfile)
            if logfile:
                click.echo('\nLog written to {}'.format(logfile))

        if with_tests:
            click.echo('\nRunning notebook tests over: {}'.format(outpath))
            if not logfile: logfile = 'tests.log'
            _notebookTest(outpath, logfile )
            click.echo('\nLog written to {}'.format(logfile))
            
        if zip:
            print('\nZipping into: {}/nYou may also want to delete the working directory ({}).'.format(repository, outpath) )
            zipper(outpath,repository)
        else:
            print('\n\nTo zip the downloaded directory, run something like: {}'.format('tm351zip {o} {z}\n\nTo run a notebook processor (OPTIONS: runWithErrors, clearOutput) while zipping: tm351zip "{o}" {z} --file-processor OPTION\n'.format(o=outpath,z=repository.name)))

        directory=''
        another = click.prompt('\Download another directory from this branch? (To quit: -)')

     #TODO
     #print('\n\nTo run this command again: {}'.format())



import click
from pathlib import Path
import nbformat
from nbformat.v4.convert import random_cell_id

#This is a precautionary step...
# Try not to clobber files we may have created previously
import os
import warnings

def listify(item):
    ''' If presented with a string and a list is required, make a list... '''
    item = [] if item is None else item
    #We may be passed a tuple - in which case, listify...
    item = list(item) if isinstance(item,(list,tuple)) else [item]
    return item


# Merge notebooks
@click.command()
@click.option("--outfile","-o", default="merged_notebook.ipynb", type=click.Path(resolve_path=False))
@click.argument('files', nargs=-1)
def nb_merger(outfile, files):
    """Merge two or more notebooks. Note that this may overwrite a pre-existing file."""
    # Opinionated on this
    version = 4 #nbformat.NO_CONVERT
    files = listify(files)

    if len(files) < 2:
        print(f"You need to pass at least two notebook filenames in order to merge them...")
        return

    # Reading the notebooks
    notebooks = [nbformat.read(nb, version) for nb in files]

    # Create new (merged) notebook
    nb_merged = nbformat.v4.new_notebook(metadata=notebooks[0].metadata)

    # Concatenate notebooks
    nb_merged.cells = []
    for notebook in notebooks:
        nb_merged.cells.extend(notebook.cells)

    for i, _ in enumerate(nb_merged.cells):
        nb_merged.cells[i]["id"] = random_cell_id()
 
    # Validate - exception if we fail
    nbformat.validate(nb_merged)

    # Save new notebook 
    nbformat.write(nb_merged, outfile)

# Split notebooks
@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--splitter", default="#--SPLITHERE--", help="String to split on (default: #--SPLITHERE--")
@click.option('--overwrite/--no-overwrite',default=False, help='Overwrite pre-existing files')
def nb_splitter(path, splitter, overwrite):
    """Split a notebook at a splitpoint marker.
    Note that this may overwrite pe-existing files.
    """

    f = Path(path)
    if f.is_file() and f.suffix == '.ipynb' :
            print(f"Checking {f}")
    else:
        print("I was expecting a Jupyter notebook .ipynb file...")
        return

    # Opinionated on this
    version = 4 #nbformat.NO_CONVERT
    nb = nbformat.read(path, as_version=version)

    #Make a copy of the notebook, just in case...
    nb2 = nb.copy()

    #We are going to see if we can split the notebook into separate parts
    parts=[]

    #Each part will contain the cells for the part
    #The rest of the notebook structure, (notebook metadata etc) will be copied from the original notebook
    partcells = []

    #Iterate through all the cells
    for cell in nb2['cells']:
        #Check for a splitline marker - go defensive!
        splitline = cell['source'].upper().strip().replace(" ","") == splitter
        #If we're not at a split line,
        if not splitline:
            #Append the cell to the cell list for this part
            partcells.append(cell)
        else:
            #Otherwise, save the cells to the part...
            parts.append(partcells)
            #...and create a new part cells list
            partcells=[]

    #Commit the final set of cells to the final part
    parts.append(partcells)

    #We can provide more warning notices...
    caution = False

    for ix, cells in enumerate(parts):
        part_fn = path.replace('.ipynb','_PART {}.ipynb'.format(ix+1))
        if os.path.isfile(part_fn) and not overwrite:
            raise Exception('File {} already exists. Set: overwrite=True. Exiting...'.format(part_fn))
        elif caution:
            warnings.warn("You were warned... The following file will be overwritten: {}".format(part_fn))
            
        #For each part, write out a separate notebook containing just the cells in that part. 
        nb_out = nb.copy()
        nb_out['cells'] = cells
        # Validate - exception if we fail
        nbformat.validate(nb_out)
        print('Writing {}...'.format(part_fn))
        nbformat.write(nb_out, part_fn, version)




import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path
import re

# From Cluadue.ai
def _process(p, max_heading_level=3):
    """
    Process a Jupyter notebook file, splitting markdown cells at headings.

    :param p: Path to the notebook file
    :param max_heading_level: Maximum heading level to split at (default: 3)
    """
    if p.is_file() and p.suffix == ".ipynb":
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)

        new_cells = []
        updated = False
        for cell in nb.cells:
            if cell["cell_type"] == "markdown":
                source = (
                    cell["source"]
                    if isinstance(cell["source"], str)
                    else "".join(cell["source"])
                )

                # Split the source at headings
                pattern = r"^(#{1," + str(max_heading_level) + r"}\s+.+)$"
                parts = re.split(pattern, source, flags=re.MULTILINE)
                # If there are no splits, keep the cell as is
                if len(parts) == 1:
                    new_cells.append(cell)
                else:
                    updated = True
                    # The first part might not start with a heading
                    if parts[0]:
                        new_cells.append(nbformat.v4.new_markdown_cell(parts[0]))

                    # Process the rest of the parts
                    for i in range(1, len(parts), 2):
                        heading = parts[i] + " " + parts[i + 1].lstrip()
                        new_cells.append(nbformat.v4.new_markdown_cell(heading))
            else:
                new_cells.append(cell)

        if updated:
            nb.cells = new_cells
            print(f"Updating {p}")
            with p.open("w") as f:
                nbformat.write(nb, f)
        else:
            print(f"No changes needed for {p}")


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--maxheading",
    type=int,
    default=3,
    help="Maximum heading level to split at (1 to 6); default=3.",
)
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
def section_splitter(paths, maxheading, recursive):
    """Split cells at section headings."""
    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir, maxheading)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p, maxheading)
        else:
            for p in nb_dir.iterdir():
                _process(p, maxheading)



# # Simple Routine to try to ensure tag toolbar is collapsed
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path

def _process(p):
    """Handle tag toolbar collapse."""

    if p.is_file() and p.suffix == '.ipynb':
        updated = False

        # Read notebook
        with p.open('r') as f:
            #print(f"Trying {p}")
            
            nb = nbformat.read(f, nbformat.NO_CONVERT)
            if "celltoolbar" in nb.metadata and nb.metadata.celltoolbar=="Tags":
                nb.metadata.pop("celltoolbar", None)
                updated = True

        if updated:
            print(f"Updating {p}")
            nbformat.write(nb, p.open('w'), nbformat.NO_CONVERT)



@click.command()
@click.argument('path', type=click.Path(resolve_path=False))
@click.option('--recursive/--no-recursive',default=True, help='Recursive search of directories.')
def tag_toolbar_collapser(path, recursive):
    """Collapse tags toolbar."""

    # Parse notebooks
    nb_dir = Path(path)
    if nb_dir.is_file():
        _process(nb_dir)
    elif recursive:
        exclude = set([])
        for dirname, subdirs, files in os.walk(path, topdown=True):
            subdirs[:] = [d for d in subdirs if d not in exclude]
            exclude_hidden_items(subdirs)
            for p in files:
                _process(Path(dirname) / p)
    else:
        for p in nb_dir.iterdir():
            _process(p)




# # Simple Routine to remove cells with a particular tag
# either throughout or just at end of notebook
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path


def _process(p, cell_typs, tags):
    """Remove tagged cells."""

    def is_tagged(cell, cell_typs, tags):
        """Return True if cell is correct type and contains a tag in tags."""
        if cell["cell_type"] not in cell_typs:
            return False
        if "metadata" in cell and "tags" in cell["metadata"]:
            return bool(set(cell["metadata"]["tags"]) & set(tags))
        return False
    
    if p.is_file() and p.suffix == ".ipynb":
        # Read notebook
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)
            cell_count = len(nb.cells)
            nb.cells = [cell for cell in nb.cells if not is_tagged(cell, cell_typs, tags)]

        if cell_count != len(nb.cells):
            print(f"Updating {p}")
            # By default, we do not convert notebook version
            nbformat.write(nb, p.open("w"), nbformat.NO_CONVERT)


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
@click.option(
    "--blitz/--no-blitz",
    default=True,
    help="Review all cell types (false if any explicit cell types are set)",
)
@click.option(
    "--code/--no-code",
    default=False,
    help="Check code cell.",
)
@click.option(
    "--md/--no-md",
    default=False,
    help="Check markdown cell.",
)
@click.option(
    "--raw/--no-raw",
    default=False,
    help="Check raw cell.",
)
@click.option(
    "--tag",
    "-t",
    multiple=True,
    help="Specify tags. You can use this option multiple times.",
)
def remove_tagged_cell(paths, recursive, blitz, code, md, raw, tag):
    """Clean empty cells."""
    cell_typs = set()
    tags = list(tag)
    if not tags:
        print("You must enter one or more `-t mytag` arguments")
        return
    if blitz:
        cell_typs = {"code", "markdown", "raw"}
    if code:
        cell_typs.add("code")
    if md:
        cell_typs.add("markdown")
    if raw:
        cell_typs.add("raw")
    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir, cell_typs, tags)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p, cell_typs, tags)
        else:
            for p in nb_dir.iterdir():
                _process(p, cell_typs, tags)



# # Simple Routine to clear a key from metadata
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path

def _process(p, tag):
    """Handle cell figure output autotagging."""

    if p.is_file() and p.suffix == '.ipynb':
        updated = False

        # Read notebook
        with p.open('r') as f:

            nb = nbformat.read(f, nbformat.NO_CONVERT)
            for _, cell in enumerate(nb['cells']):
                if cell["cell_type"]=="code" and cell.get("outputs"):
                    figure_output = False
                    for output in cell["outputs"]:
                        if output["output_type"] == "display_data" and "image/png" in output["data"]:
                            figure_output = True
                    if figure_output:
                        if "tags" not in cell["metadata"]:
                            cell["metadata"]["tags"] = [tag]
                            updated = True
                        elif tag not in cell["metadata"]["tags"]:
                            cell["metadata"]["tags"].append(tag)
                            updated = True

        if updated:
            print(f"Updating {p}")
            nbformat.write(nb, p.open('w'), nbformat.NO_CONVERT)



@click.command()
@click.argument('path', type=click.Path(resolve_path=False))
@click.option('--tag', '-t', type=str, default="nbval-figure", help='Tag to label figure output cells.')
@click.option('--recursive/--no-recursive', default=True, help='Recursive search of directories.')
def figure_autotagger(path, tag, recursive):
    """Autotag figure output cell."""

    # Parse notebooks
    nb_dir = Path(path)
    if nb_dir.is_file():
        _process(nb_dir, tag)
    elif recursive:
        exclude = set([])
        for dirname, subdirs, files in os.walk(path, topdown=True):
            subdirs[:] = [d for d in subdirs if d not in exclude]
            exclude_hidden_items(subdirs)
            for p in files:
                _process(Path(dirname) / p, tag)
    else:
        for p in nb_dir.iterdir():
            _process(p, tag)




#Inspired by https://stackoverflow.com/a/47138762/454773

from bs4 import BeautifulSoup
import json
import click
from pathlib import Path
import nbformat

@click.command()
@click.argument('path', type=click.Path(exists=True))
def html2ipynb(path):
    """Convert notebook HTML format to ipynb."""
    # I don't understand why click isn't handling this?
    path = Path(path)
    if path.is_file() and path.suffix == '.html':
        print(f"Checking {path}")
        # Read notebook
        with path.open('r') as f:
            nb = nbformat.v4.new_notebook()

            html = f.read()
            soup = BeautifulSoup(html, 'lxml')
            
            for d in soup.findAll("div"):
                if 'class' in d.attrs.keys():
                    for clas in d.attrs["class"]:
                        if clas in ["text_cell_render", "input_area"]:
                            # code cell
                            if clas == "input_area":
                                cell = nbformat.v4.new_code_cell(d.get_text())
                                nb.cells.append(cell)

                            else:
                                cell = nbformat.v4.new_code_cell(d.decode_contents())
                                nb.cells.append(cell)

    
    outpath = path.with_suffix('.ipynb')
    nbformat.write(nb, outpath.open('w'))


## Simple Routine to handle line separator (--) at start of cell
# Either by replacing it with a valid spearator, or remove it)

import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path
import re
from .utils import defensive_open

def _process(p, retain):
    """Handle cells and clean separators as required."""
    try:
        if p.is_file() and p.suffix == ".ipynb":
            updated = False
 
            # Read notebook
            with defensive_open(p) as f:
                nb = nbformat.read(f, nbformat.NO_CONVERT)
                for cell in nb["cells"]:

                    if cell["cell_type"] == "markdown":
                        source = cell["source"]
                        if isinstance(source, list):
                            source = "".join(source)
                        pattern = r"^-+\s*\s*(?:\n|$)"
                        needs_repair = re.match(pattern, source)
                        if bool(needs_repair):
                            if retain:
                                new_source = f"\n{source}"
                            else:
                                pattern = r"^-+\s*\n?"
                                match = re.search(pattern, source)
                                new_source = source[match.end() :]

                            # Update the cell source
                            if isinstance(cell["source"], list):
                                cell["source"] = [new_source]
                            else:
                                cell["source"] = new_source

                            updated = True
            if updated:
                print(f"Updating {p}")
                nbformat.write(nb, p.open("w"), nbformat.NO_CONVERT)
    except Exception as e:
        print(f"Error trying to clean separator in {p}: {str(e)}")

@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--retain/--no-retain",
    default=True,
    help="Retain (fix) separator or delete it.",
)
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
def separator_cleaner(paths, retain, recursive):
    """Clean separators at start of cell."""
    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir, retain)

        if recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p, retain)
        else:
            for p in nb_dir.iterdir():
                _process(p, retain)



# # Simple Routine to try to ensure headings are collapsed
#
# The collapsible headings extension will automatically collapse items below a heading,
# up to a heading of the same or higher level, if a cell has a
# `heading_collapsed=true` metadata element.
#
# Operationally, the extension then adds a `hidden=true` cell metadata element.
# # Whilst we don't need to add this, it is perhaps useful if the collapsible heading extension / template
# is not available, e.g. for Jupyter Book.

import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path


def _process(p, cnb, possibles):
    """Handle cell collapse."""

    def normalize_source(source):
        if isinstance(source, list):
            return "".join(source)
        return source

    collapse_metadata = "heading_collapsed" if cnb else "jp-MarkdownHeadingCollapsed"
    tag_set = {"style-activity", "style_activity"}
    metadata_set = {"activity"}

    if p.is_file() and p.suffix == ".ipynb":
        updated = False

        # Read notebook
        with p.open("r") as f:
            # print(f"Trying {p}")

            nb = nbformat.read(f, nbformat.NO_CONVERT)

            answer_block = False
            answer_header = False
            # header_level = 0
            # Enumerate through cells
            for i, cell in enumerate(nb["cells"]):
                cell_tags = cell.get("metadata", {}).get("tags", [])
                cell_metadata = cell.get("metadata", {})
                if tag_set.intersection(set(cell_tags)) or metadata_set.intersection(
                    set(cell_metadata)
                ):
                    if cell["cell_type"] == "markdown":
                        normalized_source = normalize_source(cell["source"]).strip()
                        if normalized_source.startswith("#"):
                            if any(
                                ans in normalized_source.lower() for ans in possibles
                            ):
                                answer_header = True
                elif "precollapse" in cell_tags and normalize_source(
                    cell["source"]
                ).strip().startswith("#"):
                    answer_header = True
                else:
                    answer_block = False
                    answer_header = False
                    # header_level = 0

                # For all cells - collapse if answer
                if answer_header:
                    updated = True
                    answer_header = False
                    answer_block = True
                    cell["metadata"][collapse_metadata] = True
                # Whilst there is a metadata flag to show a cell is hidden
                # this is automatically set by the collapsible headings extension
                # So the following is redundant, though may be useful if the collapsible
                # heading extentsion / template is not available, e.g. for Jupyter Book
                elif answer_block and cnb:
                    cell["metadata"]["hidden"] = True

        if updated:
            print(f"Updating {p}")
            nbformat.write(nb, p.open("w"), nbformat.NO_CONVERT)


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
@click.option(
    "--cnb/--no-cnb",
    default=False,
    help="Use classic notebook extension metadata value (default: use no-cnb (JupyterLab/nb7) format).",
)
@click.option(
    "--additional-possibles",
    "-ap",
    multiple=True,
    help="Additional possible strings to match. Use quotes; add separate flag per item.",
)
def activity_collapser(paths, recursive, cnb, additional_possibles):
    """Collapse activity answers."""
    possibles = ["# Our solution", "# Answer", "click on the triangle symbol"]
    possibles.extend(additional_possibles)
    possibles = [p.lower() for p in possibles]
    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(
                nb_dir,
                cnb,
                possibles=possibles,
            )
        if recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(
                        Path(dirname) / p,
                        cnb,
                        possibles=possibles,
                    )
        else:
            for p in nb_dir.iterdir():
                _process(
                    p,
                    cnb,
                    possibles=possibles,
                )



from setuptools import setup
import os

# Read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tm351-nb-utils",
    version="0.0.33",
    long_description=long_description,  # Use README content as long description
    long_description_content_type="text/markdown",
    py_modules=["tm351_nb_utils", "updaters", "html2nb", "splitmerge", "collapser"],
    packages=["ou_nb_workflow_tools"],
    install_requires=[
        "Click",
        "tabulate",
        "pytest",
        "nbval",
        "PyGithub",
        "beautifulsoup4",
        "humanize",
    ],
    entry_points="""
        [console_scripts]
        tm351zip=tm351_nb_utils:cli_zip
        tm351unzip=tm351_nb_utils:cli_unzip
        tm351zipview=tm351_nb_utils:cli_zipview
        tm351gitrepos=tm351_nb_utils:cli_gitrepos
        tm351nbtest=tm351_nb_utils:cli_nbtest
        tm351nbrun=tm351_nb_utils:cli_nbrun
        upgrade_empinken_tags = updaters:upgrade_empinken_tags
        html2ipynb = html2nb:html2ipynb
        nb_merge = splitmerge:nb_merger
        nb_split = splitmerge:nb_splitter
        nb_collapse_activities = collapser:activity_collapser
        nb_collapse_tagstoolbar = ou_nb_workflow_tools.tags_collapser:tag_toolbar_collapser
        nb_image_table = ou_nb_workflow_tools.cli:imagetable_generate
        nb_cell_metadata_strip = ou_nb_workflow_tools.metadata_cleaner:cell_metadata_key_cleaner
        nb_cell_figure_tagger = ou_nb_workflow_tools.figure_cell_autotagger:figure_autotagger
        nb_cell_separator_fixer = ou_nb_workflow_tools.separator_fixer:separator_cleaner
        cnb_collapse_head_migrate = ou_nb_workflow_tools.cnb2jl_ext_migrate:fix_collapse_headings_metadata
        nb_cell_stderr_finder = ou_nb_workflow_tools.warning_spotter:stderr_checker
        nb_empty_cell_cleaner = ou_nb_workflow_tools.blank_cell_cleaner:empty_cell_cleaner
        nb_split_sections = ou_nb_workflow_tools.section_splitter:section_splitter
        nb_cell_run_status = ou_nb_workflow_tools.cell_run_status_checker:cell_run_status_checker
        nb_html2myst_updater = ou_nb_workflow_tools.html2myst_converter:html_tag_replacer
        nb_remove_tagged_cell = ou_nb_workflow_tools.tagged_cell_remover:remove_tagged_cell
   """,
)



# # Simple Routine to try to find cell output errors, warnings etc
import os
from tm351_nb_utils import exclude_hidden_items

import click
import os

import nbformat
from pathlib import Path


def _process(p):
    """Spot warning."""
    def _get_warnings(p, nb):
        _trapped_warnings = []
        _warnings = []
        _i = 0 # Code cell count
        for i, cell in enumerate(nb["cells"]):
            if cell['cell_type']=="code":
                _i = _i+1
            if "outputs" in cell:
                for output in cell["outputs"]:
                    if "name" in output and output["name"] == "stderr":
                        msg = output["text"].split("\n")[0]
                        if "tags" in cell["metadata"] and 'raises-exception' in cell["metadata"]["tags"]:
                                _trapped_warnings.append( ( _i, msg) )
                        else:
                            _warnings.append(
                                (_i,
                                    msg,
                                )
                            )
                    elif "output_type" in output and output["output_type"]=="error":
                        
                        msg = output["ename"] if "ename" in output else "???"

                        # _warnings.append(
                        #        (
                        #            _i,
                        #            msg,
                        #        ))
                        if "tags" in cell["metadata"] and "raises-exception" in cell["metadata"]["tags"]:
                                _trapped_warnings.append((_i, msg))
                                print("appended t")
                        else:
                            _warnings.append(
                                (
                                    _i,
                                    msg,
                                )
                            )
        return _warnings, _trapped_warnings

    warnings = []
    trapped_warnings = []

    if p.is_file() and p.suffix == ".ipynb":
        # Read notebook
        with p.open("r") as f:
            nb = nbformat.read(f, nbformat.NO_CONVERT)
            _warnings, _trapped_warnings = _get_warnings(p, nb)
            warnings.extend(_warnings)
            trapped_warnings.extend(_trapped_warnings)
    if trapped_warnings or warnings:
        print(f"\n{p}\n")
        if trapped_warnings:
            print(f"Trapped warnings/errors:")
            for w in trapped_warnings:
                print(w)
            print("\n")
        if warnings:
            print(f"****Untrapped warnings/errors:****")
            for w in warnings:
                print(w)
            print("\n")


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(resolve_path=False))
@click.option(
    "--recursive/--no-recursive", default=True, help="Recursive search of directories."
)
def stderr_checker(paths, recursive):
    """Spot stderr errors and warnings."""

    for path in paths:
        # Parse notebooks
        nb_dir = Path(path)
        if nb_dir.is_file():
            _process(nb_dir)
        elif recursive:
            exclude = set([])
            for dirname, subdirs, files in os.walk(path, topdown=True):
                subdirs[:] = [d for d in subdirs if d not in exclude]
                exclude_hidden_items(subdirs)
                for p in files:
                    _process(Path(dirname) / p)
        else:
            for p in nb_dir.iterdir():
                _process(p)



# # Image Table
#
# Script to generate a gallery table showing images used in documents by folder.
#
# Table structure:
#
# - path
# - directory
# - file
# - img_path
# - alt text
# - image

#https://github.com/astanin/python-tabulate
# #%pip install tabulate
from tabulate import tabulate

# ## Recipe
#
# - iterate through subdirs of specified path
# - grab markdown files in subdir
# - generate table from images in markdown file

# First, let's create a function that will generate a list of images from a markdown file, given the path to the markdown file.
#
# Images in markdown are encoded as `![ALT_TEXT](IMAGE_FILEPATH)` or explicilty in HTML `<img />` tags. Either way, we should be able to parse them out because the approach I'm taking is to parse the markdown to HTML, then extract the `img` tags.
#
# The function will return a list of two tuples containing the image path and any alt text.

import markdown
from bs4 import BeautifulSoup

def get_images_from_md(fpath):
    """Get images from file."""
    images = []
    txt = open(fpath, 'r').read()
    html = markdown.markdown(txt)#,extensions=["markdown_image_caption.plugin"])
    soup = BeautifulSoup(html)
    for imgtag in soup.find_all('img'):
        if 'alt' not in imgtag:
            imgtag['alt'] = ''
        images.append((imgtag['src'], imgtag['alt']))
    return images

#The next step is to iterate through a set of files
import os

def parse_file_collection(rootdir='.'):
    """Parse all the markdown files in all the subdirs of a specified directory."""
    biglist = []
    for subdir, dirs, files in os.walk(rootdir):
        if '.ipynb_checkpoints' in subdir:
            continue
        for file in [f for f in files if f.endswith('md')]:
            fpath = os.path.join(subdir, file)
            _images = get_images_from_md(fpath)
            for (src, alt) in _images:
                if alt:
                    print(alt)
                img_path = os.path.join(subdir, src)
                biglist.append([alt, f'![{alt}]({img_path})', img_path.replace(rootdir, ''),
                                subdir, file, f'[{fpath}]({fpath})'])
    return biglist

# rootdir = '../content'
#biglist = parse_file_collection(rootdir)
#biglist[:2]

def write_gallery_file(biglist, outfilepath='gallery.md', headers=None, typ='github' ):
    """Generate a gallery file."""
    with open(outfilepath, 'w') as f:
        f.write(tabulate(biglist, headers=headers, tablefmt=typ))

#headers = ['Alt text', 'Image', 'Image Path', 'Path', 'Directory', 'Filename']
#write_gallery_file(biglist, headers=headers)

# !head gallery.md


import click
import ou_nb_workflow_tools.imagetable as imagetable

@click.command()
@click.argument('path', default='.', type=click.Path(exists=True))
@click.option('--out_file', '-o', default='gallery.md',  help='Gallery filename')
@click.option('--out_table_type', '-t', default='github',  help='Output table type')
def imagetable_generate(path, out_file, out_table_type):
    """Generate gallery."""
    click.echo(f'Using path: {path}')
    click.echo(f'Writing {out_table_type}-formatted gallery to: {out_file}')
    biglist = imagetable.parse_file_collection(path)
    headers = ['Alt text', 'Image', 'Image Path', 'Path', 'Directory', 'Filename']
    imagetable.write_gallery_file(biglist, headers=headers, typ=out_table_type)
